#include "seleneGuideController.h"
#include "turboPmacController.h"
#include <epicsExport.h>
#include <errlog.h>
#include <iocsh.h>

/**
 * @brief Construct a new seleneGuideController object
 *
 * @param portName              See documentation of sinqController
 * @param ipPortConfigName      See documentation of sinqController
 * @param numAxes               See documentation of sinqController
 * @param movingPollPeriod      See documentation of sinqController
 * @param idlePollPeriod        See documentation of sinqController
 * @param comTimeout  Time after which a communication timeout error
 * is declared in writeRead (in seconds)
 * @param extraParams           See documentation of sinqController
 */
seleneGuideController::seleneGuideController(
    const char *portName, const char *ipPortConfigName, int numAxes,
    double movingPollPeriod, double idlePollPeriod, double comTimeout)
    : turboPmacController(portName, ipPortConfigName, numAxes, movingPollPeriod,
                          idlePollPeriod, NUM_seleneGuide_DRIVER_PARAMS)

{

    asynStatus status = asynSuccess;

    status = createParam("MOTOR_ABSOLUTE_POSITION_RBV", asynParamFloat64,
                         &motorAbsolutePositionRBV_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "Controller \"%s\" => %s, line %d\nFATAL ERROR (creating a "
                  "parameter failed with %s).\nTerminating IOC",
                  portName, __PRETTY_FUNCTION__, __LINE__,
                  stringifyAsynStatus(status));
        exit(-1);
    }

    status = createParam("NORMALIZE", asynParamInt32, &normalize_);
    if (status != asynSuccess) {
        asynPrint(this->pasynUserSelf, ASYN_TRACE_ERROR,
                  "Controller \"%s\" => %s, line %d\nFATAL ERROR (creating a "
                  "parameter failed with %s).\nTerminating IOC",
                  portName, __PRETTY_FUNCTION__, __LINE__,
                  stringifyAsynStatus(status));
        exit(-1);
    }
}

asynStatus seleneGuideController::writeInt32(asynUser *pasynUser,
                                             epicsInt32 value) {
    int function = pasynUser->reason;

    // =====================================================================

    asynMotorAxis *axis = getAxis(pasynUser);

    // Check if the provided axis is one of the special selene guide axes
    if (function == normalize_) {

        // During initialization of an IOC, the "normalize" PV is initialized to
        // zero, which triggers this function. We don't want to call
        // normalizePositions() during IOC initialization.
        if (value != 0) {
            seleneAngleAxis *aAxis = dynamic_cast<seleneAngleAxis *>(axis);
            if (aAxis != nullptr) {
                return aAxis->normalizePositions();
            }

            seleneLiftAxis *lAxis = dynamic_cast<seleneLiftAxis *>(axis);
            if (lAxis != nullptr) {
                return lAxis->normalizePositions();
            }

            seleneOffsetAxis *oAxis = dynamic_cast<seleneOffsetAxis *>(axis);
            if (oAxis != nullptr) {
                return oAxis->normalizePositions();
            }
        }

        // Any other axis should just ignore an input to "Normalize"
        return asynSuccess;
    }
    return turboPmacController::writeInt32(pasynUser, value);
}

/*************************************************************************************/
/** The following functions are C-wrappers, and can be called directly from
 * iocsh */

extern "C" {

/*
C wrapper for the controller constructor. Please refer to the
seleneGuideController constructor documentation.
*/
asynStatus seleneGuideCreateController(const char *portName,
                                       const char *ipPortConfigName,
                                       int numAxes, double movingPollPeriod,
                                       double idlePollPeriod,
                                       double comTimeout) {
    /*
    We create a new instance of the controller, using the "new" keyword to
    allocate it on the heap while avoiding RAII.
    https://github.com/epics-modules/motor/blob/master/motorApp/MotorSrc/asynMotorController.cpp
    https://github.com/epics-modules/asyn/blob/master/asyn/asynPortDriver/asynPortDriver.cpp

    The created object is registered in EPICS in its constructor and can safely
    be "leaked" here.
    */
#pragma GCC diagnostic ignored "-Wunused-but-set-variable"
#pragma GCC diagnostic ignored "-Wunused-variable"
    seleneGuideController *pController =
        new seleneGuideController(portName, ipPortConfigName, numAxes,
                                  movingPollPeriod, idlePollPeriod, comTimeout);

    return asynSuccess;
}

/*
Define name and type of the arguments for the CreateController function
in the iocsh. This is done by creating structs with the argument names and
types and then providing "factory" functions
(configCreateControllerCallFunc). These factory functions are used to
register the constructors during compilation.
*/
static const iocshArg CreateControllerArg0 = {"Controller name (e.g. mcu1)",
                                              iocshArgString};
static const iocshArg CreateControllerArg1 = {"Asyn IP port name (e.g. pmcu1)",
                                              iocshArgString};
static const iocshArg CreateControllerArg2 = {"Number of axes", iocshArgInt};
static const iocshArg CreateControllerArg3 = {"Moving poll rate (s)",
                                              iocshArgDouble};
static const iocshArg CreateControllerArg4 = {"Idle poll rate (s)",
                                              iocshArgDouble};
static const iocshArg CreateControllerArg5 = {"Communication timeout (s)",
                                              iocshArgDouble};
static const iocshArg *const CreateControllerArgs[] = {
    &CreateControllerArg0, &CreateControllerArg1, &CreateControllerArg2,
    &CreateControllerArg3, &CreateControllerArg4, &CreateControllerArg5};
static const iocshFuncDef configSeleneGuideCreateController = {
    "seleneGuideController", 6, CreateControllerArgs};
static void configSeleneGuideCreateControllerCallFunc(const iocshArgBuf *args) {
    seleneGuideCreateController(args[0].sval, args[1].sval, args[2].ival,
                                args[3].dval, args[4].dval, args[5].dval);
}

// This function is made known to EPICS in turboPmac.dbd and is called by EPICS
// in order to register both functions in the IOC shell
static void seleneGuideControllerRegister(void) {
    iocshRegister(&configSeleneGuideCreateController,
                  configSeleneGuideCreateControllerCallFunc);
}
epicsExportRegistrar(seleneGuideControllerRegister);

} // extern "C"
